/*  plumb - the pipe shell - pipe helper
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "plumb.h"
#include "pipe.h"

pipe_t *pipes_head = NULL;



static char *print_pipes(fd_t *f1, fd_t *f2, fd_t *target)
{
	static char *buff = NULL;
	static int alloced = 0;
	int len;
	char *end;

	len = strlen(f1->proc.v->h.id) + strlen(f2->proc.v->h.id) + 32;
	if (target != NULL)
		len += strlen(target->proc.v->h.id) + 32;
	if (len >= alloced) {
		alloced = len;
		buff = realloc(buff, alloced);
	}

	end = buff + sprintf(buff, "%s:%d and %s:%d",
	                            f1->proc.v->h.id, f1->fd,
	                            f2->proc.v->h.id, f2->fd);
	if (target != NULL)
		sprintf(end, " connecting to %s:%d", target->proc.v->h.id, target->fd);
	return buff;
}

static void chk_multi(fd_t *f1, fd_t *f2, fd_t *target)
{
	if ((f1 != f2) && (f1 != NULL))
		plumb_error("multiple pipes trying to claim the same fd: %s\n", print_pipes(f1, f2, target));
}

/* notify virtual processes about new fds
   for real processes allow it only in stopped mode or if explicitly allowed */
#define new_fd(f) \
	do { \
		if (f->proc.v->h.virt) { \
			if (f->proc.v->newfd != NULL) \
				f->proc.v->newfd(f->proc.v, f->fd); \
		} \
		else if ((!f->proc.r->accept_fd) && (f->proc.v->h.state != PS_STOPPED))\
			plumb_error("Can not bind new pipes to real process %s\n", f->proc.v->h.id); \
	} while(0)


int pipe_register(fd_t *from, fd_t *to)
{
	pipe_t *p = NULL;
	int new_from, new_to;

	/* check if from fd already has a pipe and if the ->from of that pipe is our from fd */
	if (from->pipe != NULL) {
		chk_multi(from->pipe->from, from, to);
		chk_multi(from->pipe->to, to, from);
		p = from->pipe;
		new_from = 0;
	}
	else
		new_from = 1;

	/* check if to fd already has a pipe and if the ->to of that pipe is our to fd
	   also check if both from and to had a pipe, that's the very same pipe */
	if (to->pipe != NULL) {
		chk_multi(to->pipe->to, to, from);
		chk_multi(to->pipe->from, from, to);
		if ((p != NULL) && (p != to->pipe)) {
			plumb_error("trying to connect 2 pipes?! %s\n", print_pipes(from, to, NULL));
		}
		p = to->pipe;
		new_to = 0;
	}
	else
		new_to = 1;

	/* allocate pipe if we couldn't resuse an existing one */
	if (p == NULL) {
		p = malloc(sizeof(pipe_t));
		p->next = pipes_head;
		p->prev = NULL;
		if (pipes_head != NULL)
			pipes_head->prev = p;
		pipes_head = p;
	}

	/* make pipe -> fd and fd -> pipe bindings */
	p->from = from;
	p->to = to;

	from->pipe = p;
	to->pipe = p;

	if (new_from)
		new_fd(from);
	if (new_to)
		new_fd(to);

	return 0;
}

#define del_fd(f) \
	do { \
		if ((f->proc.v->h.virt) && (f->proc.v->delfd != NULL)) \
			f->proc.v->delfd(f->proc.v, f->fd); \
	} while(0)

#warning TODO: mark real process' fd unbound to avoid polling
void pipe_kill(pipe_t *pipe)
{
	/* detach from FDs */
	if (pipe->from != NULL) {
		del_fd(pipe->from);
		pipe->from->pipe = NULL;
		pipe->from = NULL;
	}
	if (pipe->to != NULL) {
		del_fd(pipe->to);
		pipe->to->pipe = NULL;
		pipe->to = NULL;
	}

	/* unlink from the list */
	if (pipe->prev != NULL)
		pipe->prev->next = pipe->next;
	else
		pipes_head = pipe->next;

	if (pipe->next != NULL)
		pipe->next->prev = pipe->prev;

	free(pipe);
}

void pipe_rebind(pipe_t *pipe, fd_t *src, fd_t *sink)
{
	/* one end must be the same */
	assert((src == pipe->from) || (sink == pipe->to));

	if (src != pipe->from) {
		pipe->from->pipe = NULL;
		pipe->from = src;
		if (src->pipe != NULL)
			src->pipe->from = NULL;
		src->pipe = pipe;
	}

	if (sink != pipe->to) {
		pipe->to->pipe = NULL;
		pipe->to = sink;
		if (sink->pipe != NULL)
			sink->pipe->to = NULL;
		sink->pipe = pipe;
	}
}


void pipe_uninit(void)
{
	pipe_t *p, *next;
	for(p = pipes_head; p != NULL; p = next) {
		next = p->next;
		free(p);
	}
}
