/*  plumb - the pipe shell - command line argument parsing and main loop
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "libporty/host/fs.h"
#include "parse_common.h"
#include "process.h"
#include "vp.h"
#include "plumb_gram.tab.h"
#include "config.h"
#include "diagnostic.h"
#include "timer.h"
#include "var.h"
#include "stdscript.h"

static const char *default_lib_search[] = {
	"/usr/lib/plumb",
	"~/.plumb",
	".",
	NULL
};
int lib_search_longest = -1;
kvlist_t *custom_lib_search = NULL;

int plumb_shutdown = 0;

void yyerror(void *loc, void *scanner, const char *s)
{
	YYLTYPE *yylloc = loc;
	parser_t *parser = scanner;
	plumb_error("syntax error: %s at '%s':%d:%d\n", s, parser->inp_name, yylloc->first_line, yylloc->first_column);
}

int yywrap(yyscan_t yyscanner) {
	return 1;
}

fd_t **wfds = NULL; /* only real fds (I/O of real processes) */
struct P_pollfd *pfds = NULL;
int wfds_alloced = 0, wfds_used = 0;
static int ever_tried_to_read = 0;

fd_t *fd_alloc(void)
{
	fd_t *fd;
	fd = malloc(sizeof(fd_t));

	fd->proc.v = NULL;
	fd->pipe = NULL;
	fd->pi = -1;
	return fd;
}

static void fd_grow(fd_t *fd)
{
	if (wfds_used >= wfds_alloced) {
		wfds_alloced += 128;
		wfds = realloc(wfds, sizeof(fd_t *) * wfds_alloced);
		pfds = realloc(pfds, sizeof(struct P_pollfd) * wfds_alloced);
	}
	fd->pi = wfds_used;
	wfds_used++;
	pfds[fd->pi].fd = -1;
	pfds[fd->pi].events = 0;
	pfds[fd->pi].revents = 0;
	wfds[fd->pi] = fd;
}

void fd_add_watch(fd_t *fd, int events)
{
	if (fd->pi < 0)
		fd_grow(fd);

	if (events & P_POLLIN)
		ever_tried_to_read = 1;

	pfds[fd->pi].events |= events;
}

void fd_del_watch(fd_t *fd, int events)
{
	if (fd->pi < 0)
		fd_grow(fd);

	pfds[fd->pi].events &= ~events;
}

int fd_query_watch(fd_t *fd, int events)
{
	if (fd->pi < 0)
		fd_grow(fd);

	return pfds[fd->pi].events & events;
}

void fd_set_socket(fd_t *fd, P_net_socket sockfd)
{
	if (fd->pi < 0)
		fd_grow(fd);
	pfds[fd->pi].fd = sockfd;
}

static void fd_free_all()
{
	if (wfds != NULL) {
		free(wfds);
		wfds = NULL;
	}
	if (pfds != NULL) {
		free(pfds);
		pfds = NULL;
	}
}

P_net_socket fd_query_socket(fd_t *fd)
{
	if (fd->pi < 0)
		fd_grow(fd);

	return pfds[fd->pi].fd;
}

/* remove an fd from the watch list */
static void fd_remove(fd_t *fd) {
	if (fd->pi < wfds_used) {
		memmove(pfds + fd->pi, pfds + fd->pi + 1, (wfds_used - fd->pi) * sizeof(struct P_pollfd));
		memmove(wfds + fd->pi, wfds + fd->pi + 1, (wfds_used - fd->pi) * sizeof(fd_t *));
	}
	wfds_used--;
	fd->pi = -1;
}

void do_read(P_net_socket sockfd, fd_t *fd, int hup_mode)
{
	char buff[READ_BUF_SIZE];
	int len;

#ifdef RW_DEBUG
	len = P_net_read(sockfd, buff, sizeof(buff));
#else
	len = P_net_read(sockfd, buff, sizeof(buff)-1);
#endif

#ifdef RW_DEBUG
		buff[len] = '\0';
		fprintf(stderr, "read: [%d]: %d: ", sockfd, len);
		fprintstr(stderr, buff);
		fprintf(stderr, "\n");
#endif


	if (len < 0) {       /* handle errors */
		if (P_net_inprogress)   /* this should not happen with nonblocking sockets, but handle it just in case */
			return;
		fd_close(fd);
		fd_del_watch(fd, P_POLLIN);
		len = 0;
	}

	/* HUP-mode read needs special treatment */
	if (hup_mode) {
		read_buf_t *last, *r;
		/* obuff is not null, have to cache reads, try keeping boundaries */
			/* allocate a new item */
		r = malloc(sizeof(read_buf_t) + len);
		r->len = len;
		r->next = NULL;
		if (len > 0)
			memcpy(r->data, buff, len);
		else
			fd->to_delete = 1; /* make sure HUP'd read-fd is not staying in watch after all data has been read to eof */

		/* link it to the end of the read buffer list */
		if (fd->rb != NULL) {
			for(last = fd->rb; last->next != NULL; last = last->next) ;
			last->next = r;
		}
		else
			fd->rb = r;
	}
	else {
		/* normal read */
		if (len == 0) {           /* close target on eof */
			fd_close(fd);
			fd_del_watch(fd, P_POLLIN);
		}
		else                      /* pass on, target fd is ready to accept if we got here */
			fd_write(fd, buff, len);
	}
}

static void do_write(P_net_socket sockfd, fd_t *fd)
{
	char *buff;
	int len;

	if (fd->obuff != NULL) {
		/* save buffer for sending and fd back to normal operation */
		buff = fd->obuff;
		len = fd->obuff_len;
		fd_del_watch(fd, P_POLLOUT);
		fd->obuff = NULL;
		fd->obuff_len = 0;
		/* try to send out the buffer as normal data */
		fd_write(fd, buff, len);
		free(buff);

		/* if it failed (partially or fully), it's on P_POLLOUT again and
		   obuff is set up and it is still blocking. Else it is not blocking,
		   back-promote the block release */
		if (fd->obuff == NULL) {
			back_block(fd, 0);
		}
	}
	else {
		plumb_error("Got POLLOUT for a socket not even blocking?!");
	}
}

void close_all_fds()
{
	int n;
	for(n = 0; n < wfds_used; n++)
		fd_close(wfds[n]);
}

int plumb_stick = 0;
void main_poll(int timeout)
{
	int chg, n, watch_read, delcnt;

	chg = P_poll(pfds, wfds_used, timeout);

#ifdef POLL_DEBUG
	{
		int n;
		fprintf(stderr, "W: ");
		for(n=0; n < wfds_used; n++) {
			fprintf(stderr, "%s:%d %d:%d [%d] | ", wfds[n]->proc.v->h.id, wfds[n]->fd, pfds[n].events, pfds[n].revents, pfds[n].fd);
		}
		fprintf(stderr, "\n");
	}
#endif

	/* check for events on each fd */
	watch_read = 0;
	for(n = 0; n < wfds_used; n++) {
		if (pfds[n].events & P_POLLIN) {
			watch_read++;
			/* read only if both revents and events says so: it may be that within the
			   same loop an earlier fd already get a block on a sink that affected
			   the current one, blocking it (through back-promotion). In this case
			   even if poll said _before_ the loop that we could read the current
			   fd, we won't, because we wouldn't be able to write */
			if ((pfds[n].revents & P_POLLIN) || (pfds[n].revents & P_POLLHUP))
				do_read(pfds[n].fd, wfds[n], 0);
		}
		else if (pfds[n].revents & P_POLLHUP) {
			/* didn't intend to read but got HUP - have to read all in internal buffer */
			if (is_source(wfds[n]))
				do_read(pfds[n].fd, wfds[n], 1);
			else if (is_sink(wfds[n])) {
				/* HUP on a write socket */
				/*fprintf(stderr, "whup on %s:%d\n", wfds[n]->proc.v->h.id, wfds[n]->fd);*/
				fd_close(wfds[n]);
			}
			else {
				plumb_error("HUP on an fd which is neither sink nor source");
			}
		}
		if (pfds[n].revents & P_POLLOUT)
			do_write(pfds[n].fd, wfds[n]);
	}
		/* collect dead fds and remove them from the watch list - it's required
	   because setting events to 0 will still let poll() to return immediately
	   with errors; actually it's no way to convince poll() to ignore HUP. */
	delcnt = 0;
	for(n = 0; n < wfds_used; n++) {
		if (wfds[n]->to_delete) {
			int i;
			wfds[n]->to_delete = 0;
			fd_remove(wfds[n]);
			delcnt++;

			/* need to fix pollfd indices after deleting */
			if (delcnt > 0)
				for(i = n; i < wfds_used; i++)
					wfds[i]->pi = i;

			n--;
		}
	}
}

void main_loop(void)
{
	int left;

	while(plumb_stick) {
		left = timer_run();

		/* it may have changed in the timer if timer killed processes - in this case quit loop */
		if (plumb_stick <= 0)
			break;
		main_poll(left);
	}
}

void plumb_stop_RPs(void)
{
	int n, timeout = 0;
	const char *stimeout;

	/* stop standard script */
	plumb_parse_string(NULL, NULL, "standard script stop", plumb_stdstop, NULL);

	/* close standard IO of all real processes and hope they'd die */
	proc_close_stdio_all(0);
	stimeout = var_get(NULL, "plumb_kill_timeout");
	if (stimeout != NULL)
		timeout = atoi(stimeout) * 100;
	if (timeout <= 0)
		timeout = DEFAULT_PLUMB_KILL_TIMEOUT;
	for(n = 0; n < timeout; n++) {
		int r;
		r = proc_count_real_running();
		if (r == 0)
			break;
/*		fprintf(stderr, "{%d} still running: %d\n", n, r); */
		/* after half the time, terminate them */
		if (n == timeout/2)
			proc_stop_all(0, 0);

		/* run poll on the script: stdio of RPs are already closed, but the VP
		   infrastructure needs to be run to let _procwatch collect all the die
		   signals */
		main_poll(10);
	}
	/* if anything is still running after a second, kill it with fire */
	proc_stop_all(1, 1);
}

void plumb_exit(int retval)
{
	plumb_shutdown = 1;
	/* clean exit - free up resources */
	plumb_stop_RPs();
	close_all_fds();
	fd_free_all();
	timer_uninit();
	plumb_event_uninit();
	proc_free_all();
	vp_uninit();
	var_uninit(NULL);
	pipe_uninit();
	if (custom_lib_search != NULL)
		kvlist_free(custom_lib_search);
	P_net_uninit();
	exit(retval);
}

static void lib_search_init()
{
	const char **lp;
	int dlen;

	for(lp = default_lib_search; *lp != NULL; lp++) {
		dlen = strlen(*lp);
		if (dlen > lib_search_longest)
			lib_search_longest = dlen;
	}
}

static void asm_path(char *path, const char *dir, const char *fn, int len)
{
	int dlen;

	dlen = strlen(dir);
	memcpy(path, dir, dlen);
	path[dlen] = P_path_sep;
	memcpy(path+dlen+1, fn, len+1);
}

char *plumb_lib_search(const char *fn)
{
	const char **lp;
	int len;
	char *path;
	P_stat_t st;
	kv_t *k;

	len = strlen(fn);
	path = malloc(lib_search_longest + len + 8);

	for(lp = default_lib_search; *lp != NULL; lp++) {
		asm_path(path, *lp, fn, len);
		if (P_stat(path, &st) == 0)
			return path;
	}

	if (custom_lib_search != NULL) {
		for(k = custom_lib_search->head; k != NULL; k = k->next) {
			asm_path(path, k->key, fn, len);
			if (P_stat(path, &st) == 0)
				return path;
		}
	}

	free(path);
	return NULL;
}


int main(int argc, char *argv[])
{
	int n;
	FILE *inpf;
	char *inp_name;

	P_net_init();
	timer_init();
	vp_init();
	lib_search_init();
	plumb_parse_string(NULL, NULL, "plumb_stdscript", plumb_stdscript, NULL);


	for(n = 1; n < argc; n++) {
		char *cmd, *arg;
		cmd = argv[n];
		arg = argv[n+1];
		if (*cmd == '-') {
			while(*cmd == '-') cmd++;
			switch(*cmd) {
				case 'f':
					inpf = fopen(arg, "r");
					if (inpf == NULL)
						plumb_error("failed to open script %s for -f\n", arg);
					inp_name = malloc(strlen(arg) + 8);
					sprintf(inp_name, "-f %s", arg);
					plumb_parse_file(NULL, NULL, inp_name, inpf, NULL);
					free(inp_name);
					fclose(inpf);
					n++;
					break;
				case 'v':
					if (n+2 >= argc)
						plumb_error("not enough arguments for -c\n");
					var_set(NULL, argv[n+1], argv[n+2]);
					n+=2;
					break;
				case 'I':
					custom_lib_search = kvlist_append(custom_lib_search, kv(strclone(argv[n+1]), NULL));
					n++;
					break;
			}
		}
		else {
			plumb_parse_string(NULL, NULL, "cmdline", cmd, NULL);
		}
	}

	main_loop();

	plumb_exit(0);
	return 1;
}
