/*  plumb - the pipe shell - lexycal
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

%option reentrant
%option bison-bridge
%option bison-locations
%{
#define IN_LEX
#include "parse_common.h"
#include "process.h"
#include "plumb_gram.tab.h"
#include "plumb.h"
#include "var.h"
#include "config.h"

typedef struct stack_s {
	YY_BUFFER_STATE yybuff;
} pb_stack_t;

typedef struct lex_parser_s {
	pb_stack_t stack[LEX_MAX_DEPTH];
	int level;
} lex_parser_t;

#define YY_EXTRA_TYPE lex_parser_t

static int level = 0;

void lex_push();
int lex_pop();
void lex_include();

%}

idchars [A-Za-z0-9_./-]
dquote (\\\"|[^\"])
squote (\\\'|[^\'])

%%
[#].*[\r\n]     { yylloc->first_line++; yylloc->first_column = 0;}
["]{dquote}*["] { yylval->str = str_quote(yytext, '\"'); yylloc->first_column += strlen(yylval->str); return STRING; }
[']{squote}*['] { yylval->str = str_quote(yytext, '\''); yylloc->first_column += strlen(yylval->str); return STRING; }
lib-include[ ]{idchars}+[\n]  { lex_include(yyg, parser, strclone(yytext+12), 1); }
include[ ]{idchars}+[\n]      { lex_include(yyg, parser, strclone(yytext+8), 0); }
[\\]            { }
[\\][\n]        { }
[\n]            { yylloc->first_line++; yylloc->first_column = 0; return NL;}
[ \t]           { yylloc->first_column++;}
[;\r]           { yylloc->first_column++; return ';'; }
[{}\][:,=|*+]   { yylloc->first_column++; return *yytext; }
[0-9]+          { yylloc->first_column += strlen(yytext); yylval->str = strclone(yytext); return INT; }
[$]{idchars}+   { yylloc->first_column += strlen(yytext); lex_push(yyg, NULL); yy_var_subst(parser->tx, yytext, 0, yyg); }
[$][(][^\n)]*[)] { yylloc->first_column += strlen(yytext); lex_push(yyg, NULL); yy_var_subst(parser->tx, yytext, 1, yyg); }
[$][{][^\n}]*[}] { char *v; int l = strlen(yytext); yylloc->first_column += l; yytext[l-1] = '\0'; v = var_get(parser->tx, yytext+2); if (v == NULL) plumb_error("Can't find variable %s\n", yytext+2); yylval->str = strclone(v); return STRING; }

stop            { yylloc->first_column += strlen(yytext); return KW_STOP; }
kill            { yylloc->first_column += strlen(yytext); return KW_KILL; }
pause           { yylloc->first_column += strlen(yytext); return KW_PAUSE; }
resume          { yylloc->first_column += strlen(yytext); return KW_RESUME; }
flush           { yylloc->first_column += strlen(yytext); return KW_FLUSH; }
exit            { yylloc->first_column += strlen(yytext); return KW_EXIT; }
query           { yylloc->first_column += strlen(yytext); return KW_QUERY; }
dump_internals  { yylloc->first_column += strlen(yytext); return KW_DUMP_INTERNALS; }

{idchars}+      { yylloc->first_column += strlen(yytext); yylval->str = strclone(yytext); return STRING; }

.               { plumb_error("Unhandled character in input: %c (%d) - did you mean \"%c\"?\n", *yytext, *yytext, *yytext); }

<<EOF>>         { if (lex_pop(yyscanner, yyg)) { yyterminate(); } }
%%

void lex_push(struct yyguts_t *yyg)
{
	yyextra.stack[level].yybuff = YY_CURRENT_BUFFER;
	yyextra.level++;
}

int lex_pop(yyscan_t *scanner, struct yyguts_t *yyg)
{
	if (yyextra.level > 0) {
		yyextra.level--;
		yy_switch_to_buffer(yyextra.stack[level].yybuff, scanner);
		return 0;
	}
	return 1;
}

char *lex_getpos(void *scanner)
{
	struct yyguts_t *yyg = scanner;
	return yytext;
}


int lex_is_token_singlechar(int token)
{
	switch(token) {
		case NL:
			return 1;
		case INT:
		case STRING:
		case KW_STOP:
		case KW_KILL:
		case KW_PAUSE:
		case KW_RESUME:
		case KW_FLUSH:
		case KW_EXIT:
		case KW_QUERY:
		case KW_DUMP_INTERNALS:
			return 0;
		default:
			if ((token > 0) && (token < 255))
				return 1;
		plumb_error("Internal error: lex_is_token_singlechar(): unhandled token %d\n", token);
	}
	return 0;
}

void lex_include(struct yyguts_t *yyg, parser_t *p, const char *fn_, int tx_rename)
{
	transaction_t *tx = p->tx;
	char *end, *fn;

	fn = fn_;
	while(isspace(*fn)) fn++;
	end = strpbrk(fn, "\r\n");
	if (end != NULL)
		*end = '\0';

	lex_push(yyg);
	if (tx_rename) {
		int lfn, ltx;
		char *newname;

		lfn = strlen(fn);
		ltx = strlen(tx->name);
		newname = malloc(lfn + ltx + 128);
		sprintf(newname, "%s (included from %s:%d)", fn, tx->name, yylloc->first_line+1);
		tx = tx_this_or_new(NULL, newname, tx->query_out);
		exe_append_subtx(p, tx);
	}
	yy_include(tx, NULL, fn);
	yylloc->first_line++;
	yylloc->first_column = 0;
	free(fn_);
}
